@@ -22,6 +22,7 @@ import (
 	"k8s.io/autoscaler/cluster-autoscaler/config"
 	"k8s.io/autoscaler/cluster-autoscaler/estimator"
 	"k8s.io/autoscaler/cluster-autoscaler/expander"
+	processor_callbacks "k8s.io/autoscaler/cluster-autoscaler/processors/callbacks"
 	"k8s.io/autoscaler/cluster-autoscaler/simulator"
 	kube_util "k8s.io/autoscaler/cluster-autoscaler/utils/kubernetes"
 	kube_client "k8s.io/client-go/kubernetes"
@@ -45,6 +46,8 @@ type AutoscalingContext struct {
 	ExpanderStrategy expander.Strategy
 	// EstimatorBuilder is the builder function for node count estimator to be used.
 	EstimatorBuilder estimator.EstimatorBuilder
+	// ProcessorCallbacks is interface defining extra callback methods which can be called by processors used in extension points.
+	ProcessorCallbacks processor_callbacks.ProcessorCallbacks
 }
 
 // AutoscalingKubeClients contains all Kubernetes API clients,
@@ -81,7 +84,9 @@ func NewResourceLimiterFromAutoscalingOptions(options config.AutoscalingOptions)
 
 // NewAutoscalingContext returns an autoscaling context from all the necessary parameters passed via arguments
 func NewAutoscalingContext(options config.AutoscalingOptions, predicateChecker *simulator.PredicateChecker,
-	autoscalingKubeClients *AutoscalingKubeClients, cloudProvider cloudprovider.CloudProvider, expanderStrategy expander.Strategy, estimatorBuilder estimator.EstimatorBuilder) *AutoscalingContext {
+	autoscalingKubeClients *AutoscalingKubeClients, cloudProvider cloudprovider.CloudProvider,
+	expanderStrategy expander.Strategy, estimatorBuilder estimator.EstimatorBuilder,
+	processorCallbacks processor_callbacks.ProcessorCallbacks) *AutoscalingContext {
 	return &AutoscalingContext{
 		AutoscalingOptions:     options,
 		CloudProvider:          cloudProvider,
@@ -89,6 +94,7 @@ func NewAutoscalingContext(options config.AutoscalingOptions, predicateChecker *
 		PredicateChecker:       predicateChecker,
 		ExpanderStrategy:       expanderStrategy,
 		EstimatorBuilder:       estimatorBuilder,
+		ProcessorCallbacks:     processorCallbacks,
 	}
 }
 
